# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
from typing import Union
from numpy.typing import NDArray


class Particle:
    def __init__(
        self,
        mass_kg: Union[float, np.float64],
        radius_m: Union[float, np.float64],
        initial_time_s: Union[float, np.float64],
        initial_position_m: NDArray[np.float64],
        initial_velocity_m_per_s: NDArray[np.float64],
        initial_rotation_rad_per_s: NDArray[np.float64],
    ) -> None:
        self._mass_kg = np.float64(mass_kg)
        self._radius_m = np.float64(radius_m)

        self._characteristic_length_m = 2.0 * np.float64(radius_m)
        self._inertia_kg_m2 = 0.4 * self.mass_kg * self.radius_m**2

        # Initial values at t=t0
        self._initial_time_s = np.float64(initial_time_s)
        self._initial_position_m = initial_position_m.copy()
        self._initial_velocity_m_per_s = initial_velocity_m_per_s.copy()
        self._initial_rotation_rad_per_s = initial_rotation_rad_per_s.copy()

        # Initial energy
        self._initial_kinetic_energy_J = 0.5 * self._mass_kg * np.linalg.norm(self._initial_velocity_m_per_s) ** 2
        self._initial_rotation_energy_J = 0.5 * self._inertia_kg_m2 * np.linalg.norm(self._initial_rotation_rad_per_s) ** 2

    @property  # read only
    def mass_kg(self) -> np.float64:
        return self._mass_kg

    @property  # read only
    def radius_m(self) -> np.float64:
        return self._radius_m

    @property  # read only
    def characteristic_length_m(self) -> np.float64:
        return self._characteristic_length_m

    @property  # read only
    def inertia_kg_m2(self) -> np.float64:
        return self._inertia_kg_m2

    @property  # read only
    def initial_time_s(self) -> np.float64:
        return self._initial_time_s

    @property  # read only
    def initial_position_m(self) -> NDArray[np.float64]:
        return self._initial_position_m

    @property  # read only
    def initial_velocity_m_per_s(self) -> NDArray[np.float64]:
        return self._initial_velocity_m_per_s

    @property  # read only
    def initial_rotation_rad_per_s(self) -> NDArray[np.float64]:
        return self._initial_rotation_rad_per_s

    @property  # read only
    def initial_kinetic_energy_J(self) -> np.float64:
        return self._initial_kinetic_energy_J

    @property  # read only
    def initial_rotation_energy_J(self) -> np.float64:
        return self._initial_rotation_energy_J
