# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
import scipy as sp
from typing import Union
from numpy.typing import NDArray
from class_def.environment import Environment
from class_def.particle import Particle
from physical_quantities.fluid_quantities import density_kg_per_m3, reynolds_number, mach_number, drag_coefficients_sphere, skin_friction_coefficients


def eom_bb_bullet_exact(t: np.float64, x: NDArray[np.float64], p: Particle, env: Environment) -> NDArray[np.float64]:
    x_rhs = np.empty(x.shape[0], dtype=np.float64)

    # Physical parameters
    m = p.mass_kg  # mass of BB bullet
    R = p.radius_m  # radius of BB bullet
    L = p.characteristic_length_m  # characteristic_length_m
    Inertia = p.inertia_kg_m2  # Inertia of sphere

    ftype = env.fluid_type
    P = env.pressure_pa  # pressure in Pa
    T = env.temprature_celsius  # temprature in celsius
    g = env.gravity_m_per_s2  # gravitational acceleration
    phi = env.relative_humidity_percent  # relative humidity in percent
    u = env.flow_velocity_m_per_s  # flow velocity in m/s

    v = x[3:6]
    V = np.array([v[0] - u[0], v[1] - u[1], v[2] - u[2]], dtype=np.float64)
    V_norm = np.linalg.norm(V)

    w = x[6:9]
    V_cross_w = np.cross(V, w)

    rho = density_kg_per_m3(T, P, phi, ftype)
    Re = reynolds_number(V_norm, L, T, P, phi, ftype)
    Ma = mach_number(V_norm, T, P, phi, ftype)
    Cd = drag_coefficients_sphere(Re, Ma, ftype)
    Cl = 0.12
    Cf = skin_friction_coefficients(Re, ftype)

    # Force
    # - Gravity force
    Fg = np.array([0.0, 0.0, -m * g], dtype=np.float64)
    # - Drag force
    Fd = -0.5 * Cd * rho * np.pi * R * R * V_norm * V
    # - Magnus force
    Fl = -Cl * ((4.0 * np.pi * R**3) / 3.0) * 2.0 * rho * V_cross_w

    # - Moment of force
    tmp = 0.5 * Cf * rho * R**3
    N_sphere = tmp * integral_in_moment_of_force(v, w, u, R)

    # Define right-hand side of equation of motion
    x_rhs[0:3] = v[0:3]
    x_rhs[3:6] = (Fg + Fd + Fl) / m
    x_rhs[6:9] = N_sphere / Inertia

    print(t)

    return x_rhs


def integral_in_moment_of_force(v: NDArray[np.float64], w: NDArray[np.float64], u: NDArray[np.float64], R: Union[float, np.float64]) -> np.float64:
    theta_a_limits = 0.0
    theta_b_limits = np.pi
    phi_a_limits = 0.0
    phi_b_limits = 2.0 * np.pi

    result = np.empty(3, dtype=np.float64)
    error = np.empty(3, dtype=np.float64)

    epsabs_in_integral = 1e-4
    epsrel_in_integral = 1e-4

    component = "x"
    result[0], error[0] = sp.integrate.dblquad(
        integrand,
        theta_a_limits,
        theta_b_limits,
        phi_a_limits,
        phi_b_limits,
        args=(
            v,
            w,
            u,
            R,
            component,
        ),
        epsabs=epsabs_in_integral,
        epsrel=epsrel_in_integral,
    )

    component = "y"
    result[1], error[1] = sp.integrate.dblquad(
        integrand,
        theta_a_limits,
        theta_b_limits,
        phi_a_limits,
        phi_b_limits,
        args=(
            v,
            w,
            u,
            R,
            component,
        ),
        epsabs=epsabs_in_integral,
        epsrel=epsrel_in_integral,
    )

    component = "z"
    result[2], error[2] = sp.integrate.dblquad(
        integrand,
        theta_a_limits,
        theta_b_limits,
        phi_a_limits,
        phi_b_limits,
        args=(
            v,
            w,
            u,
            R,
            component,
        ),
        epsabs=epsabs_in_integral,
        epsrel=epsrel_in_integral,
    )

    return np.float64(result)


def integrand(
    phi, theta, v: NDArray[np.float64], w: NDArray[np.float64], u: NDArray[np.float64], R: Union[float, np.float64], component: str
) -> np.float64:
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)
    sin_p = np.sin(phi)
    cos_p = np.cos(phi)
    unit_vector_radius = np.array([sin_t * cos_p, sin_t * sin_p, cos_t])
    U = u - R * np.cross(w, unit_vector_radius) - v
    U_norm = np.linalg.norm(U)
    s = U_norm * (np.cross(unit_vector_radius, U)) * sin_t
    if component == "x":
        result = s[0]
    elif component == "y":
        result = s[1]
    elif component == "z":
        result = s[2]
    else:
        raise ValueError(f"Unexpected component in integrand : component={component}")

    return np.float64(result)
