# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
import scipy as sp

# from scipy.integrate import OdeResult
from typing import Tuple, Any
from numpy.typing import NDArray
from class_def.environment import Environment
from class_def.particle import Particle

from ivp_functions.exact import eom_bb_bullet_exact
from ivp_functions.approx import eom_bb_bullet_approx
from ivp_functions.approx_w_solved import eom_bb_bullet_approx_w_solved, rotation_decaying_approx_wrapper
from ivp_functions.no_rotation_decay import eom_bb_bullet_no_rotation_decay


def event(t: np.float64, x: NDArray[np.float64], p: Particle, env: Environment) -> np.float64:
    # (x, y, z) = (x[0], x[1], x[2])
    # (v_x, v_y, v_z) = (x[3], x[4], x[5])
    # (w_x, w_y, w_z) = (x[6], x[7], x[8])
    return x[2]


def solve_ivp_for_bb_bullet(
    analysis_type: str, time_sequence_s: NDArray[np.float64], particle: Particle, environment: Environment
) -> Tuple[NDArray[np.float64], NDArray[np.float64], Any]:
    ivp_abstol = 1e-5
    ivp_reltol = 1e-5

    start_time_s = time_sequence_s[0]
    max_time_s = time_sequence_s[-1]

    # Solve initial value problem
    event.terminal = True  # type: ignore

    if analysis_type == "exact":
        # Set initial conditions
        x = np.empty(9, dtype=np.float64)
        x[0:3] = particle.initial_position_m
        x[3:6] = particle.initial_velocity_m_per_s
        x[6:9] = particle.initial_rotation_rad_per_s
        sol = sp.integrate.solve_ivp(
            eom_bb_bullet_exact,
            [start_time_s, max_time_s],
            x,
            method="RK45",
            t_eval=time_sequence_s,
            args=(
                particle,
                environment,
            ),
            events=event,
            rtol=ivp_reltol,
            atol=ivp_abstol,
        )

        # Add the time and y information when the last event occurs
        if len(sol.t_events[0]) > 0:  # if event occurs
            final_t = sol.t_events[0][-1]  # time when the last event occur
            final_y = sol.y_events[0][-1]  # values when the last event occur

            # Make new array to add the event elements
            # sol_t: (0:M)
            # sol_y: (0:N, 0:M)
            # sol_add_t: (0:M+1)
            # sol_add_y: (0:N+3, 0:M+1)
            N = sol.y.shape[0]
            M = sol.y.shape[1]
            sol_add_t = np.full(M + 1, np.nan)
            sol_add_y = np.full((N, M + 1), np.nan)

            # Python slicing and copy
            # sol_add_t: (0:M) = sol_t: (0:M)
            # sol_add_y: (0:N, 0:M) = sol_y: (0:N, 0:M)
            sol_add_t[0:M] = sol.t
            sol_add_y[0:N, 0:M] = sol.y

            # Add the final element
            # sol_add_t: (M) = final_t: (0:1)
            # sol_add_y: (0:N, M) = final_y: (0:N)
            print(final_t)
            print(final_y)
            sol_add_t[M] = final_t
            sol_add_y[0:N, M] = final_y
        else:
            sol_add_t = sol.t
            sol_add_y = sol.y

    elif analysis_type == "approx":
        # Set initial conditions
        x = np.empty(9, dtype=np.float64)
        x[0:3] = particle.initial_position_m
        x[3:6] = particle.initial_velocity_m_per_s
        x[6:9] = particle.initial_rotation_rad_per_s
        sol = sp.integrate.solve_ivp(
            eom_bb_bullet_approx,
            [start_time_s, max_time_s],
            x,
            method="RK45",
            t_eval=time_sequence_s,
            args=(
                particle,
                environment,
            ),
            events=event,
            rtol=ivp_reltol,
            atol=ivp_abstol,
        )

        # Add the time and y information when the last event occurs
        if len(sol.t_events[0]) > 0:  # if event occurs
            final_t = sol.t_events[0][-1]  # time when the last event occur
            final_y = sol.y_events[0][-1]  # values when the last event occur

            # Make new array to add the event elements
            # sol_t: (0:M)
            # sol_y: (0:N, 0:M)
            # sol_add_t: (0:M+1)
            # sol_add_y: (0:N, 0:M+1)
            N = sol.y.shape[0]
            M = sol.y.shape[1]
            sol_add_t = np.full(M + 1, np.nan)
            sol_add_y = np.full((N, M + 1), np.nan)

            # Python slicing and copy
            # sol_add_t: (0:M) = sol_t: (0:M)
            # sol_add_y: (0:N, 0:M) = sol_y: (0:N, 0:M)
            sol_add_t[0:M] = sol.t
            sol_add_y[0:N, 0:M] = sol.y

            # Add the final element
            # sol_add_t: (M) = final_t: (0:1)
            # sol_add_y: (0:N, M) = final_y: (0:N)
            sol_add_t[M] = final_t
            sol_add_y[0:N, M] = final_y
        else:
            sol_add_t = sol.t
            sol_add_y = sol.y

    elif analysis_type == "approx_w_solved":
        # Set initial conditions
        x = np.empty(6, dtype=np.float64)
        x[0:3] = particle.initial_position_m
        x[3:6] = particle.initial_velocity_m_per_s
        sol = sp.integrate.solve_ivp(
            eom_bb_bullet_approx_w_solved,
            [start_time_s, max_time_s],
            x,
            method="RK45",
            t_eval=time_sequence_s,
            args=(
                particle,
                environment,
            ),
            events=event,
            rtol=ivp_reltol,
            atol=ivp_abstol,
        )

        # Add the time and y information when the last event occurs
        if len(sol.t_events[0]) > 0:  # if event occurs
            final_t = sol.t_events[0][-1]  # time when the last event occur
            final_y = sol.y_events[0][-1]  # values when the last event occur

            # Make new array to add the event elements
            # sol_t: (0:M)
            # sol_y: (0:N, 0:M)
            # sol_add_t: (0:M+1)
            # sol_add_y: (0:N+3, 0:M+1)
            N = sol.y.shape[0]
            M = sol.y.shape[1]
            sol_add_t = np.full(M + 1, np.nan)
            sol_add_y = np.full((N + 3, M + 1), np.nan)

            # Python slicing and copy
            # sol_add_t: (0:M) = sol_t: (0:M)
            # sol_add_y: (0:N, 0:M) = sol_y: (0:N, 0:M)
            sol_add_t[0:M] = sol.t
            sol_add_y[0:N, 0:M] = sol.y

            # Add the final element
            # sol_add_t: (M) = final_t: (0:1)
            # sol_add_y: (0:N, M) = final_y: (0:N)
            sol_add_t[M] = final_t
            sol_add_y[0:N, M] = final_y

            # Add the omega component to a solution
            # sol_add_y: (N:N+3, 0:M) = w_approx: (0:3, 0:M)
            for i, t in enumerate(sol_add_t):
                v = np.array([sol_add_y[3, i], sol_add_y[4, i], sol_add_y[5, i]])
                w_approx_norm = rotation_decaying_approx_wrapper(t, v, particle, environment)
                sol_add_y[N : N + 3, i] = w_approx_norm * (particle.initial_rotation_rad_per_s / np.linalg.norm(particle.initial_rotation_rad_per_s))

        else:
            # Make new array to add the event elements
            N = sol.y.shape[0]
            M = sol.y.shape[1]
            sol_add_t = np.full(M, np.nan)
            sol_add_y = np.full((N + 3, M), np.nan)

            # Python slicing and copy
            sol_add_t[0:M] = sol.t
            sol_add_y[0:N, 0:M] = sol.y

            # Add the omega component to a solution
            for i, t in enumerate(sol_add_t):
                v = np.array([sol_add_y[3, i], sol_add_y[4, i], sol_add_y[5, i]])
                w_approx_norm = rotation_decaying_approx_wrapper(t, v, particle, environment)
                sol_add_y[N : N + 3, i] = w_approx_norm * (particle.initial_rotation_rad_per_s / np.linalg.norm(particle.initial_rotation_rad_per_s))

    elif analysis_type == "no_rotation_decay":
        # Set initial conditions
        x = np.empty(9, dtype=np.float64)
        x[0:3] = particle.initial_position_m
        x[3:6] = particle.initial_velocity_m_per_s
        x[6:9] = particle.initial_rotation_rad_per_s
        sol = sp.integrate.solve_ivp(
            eom_bb_bullet_no_rotation_decay,
            [start_time_s, max_time_s],
            x,
            method="RK45",
            t_eval=time_sequence_s,
            args=(
                particle,
                environment,
            ),
            events=event,
            rtol=ivp_reltol,
            atol=ivp_abstol,
        )

        # Add the time and y information when the last event occurs
        if len(sol.t_events[0]) > 0:  # if event occurs
            final_t = sol.t_events[0][-1]  # time when the last event occur
            final_y = sol.y_events[0][-1]  # values when the last event occur

            # Make new array to add the event elements
            # sol_t: (0:M)
            # sol_y: (0:N, 0:M)
            # sol_add_t: (0:M+1)
            # sol_add_y: (0:N+3, 0:M+1)
            N = sol.y.shape[0]
            M = sol.y.shape[1]
            sol_add_t = np.full(M + 1, np.nan)
            sol_add_y = np.full((N, M + 1), np.nan)

            # Python slicing and copy
            # sol_add_t: (0:M) = sol_t: (0:M)
            # sol_add_y: (0:N, 0:M) = sol_y: (0:N, 0:M)
            sol_add_t[0:M] = sol.t
            sol_add_y[0:N, 0:M] = sol.y

            # Add the final element
            # sol_add_t: (M) = final_t: (0:1)
            # sol_add_y: (0:N, M) = final_y: (0:N)
            sol_add_t[M] = final_t
            sol_add_y[0:N, M] = final_y
        else:
            sol_add_t = sol.t
            sol_add_y = sol.y
    else:
        raise ValueError(f"Unexpected component: analysis_type={analysis_type}")

    return sol_add_t, sol_add_y, sol
