# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
from numpy.typing import NDArray
from class_def.environment import Environment
from class_def.particle import Particle
from physical_quantities.fluid_quantities import density_kg_per_m3, reynolds_number, mach_number, drag_coefficients_sphere


def eom_bb_bullet_no_rotation_decay(t: np.float64, x: NDArray[np.float64], p: Particle, env: Environment) -> NDArray[np.float64]:
    x_rhs = np.empty(x.shape[0], dtype=np.float64)

    # Physical parameters
    m = p.mass_kg  # mass of BB bullet
    R = p.radius_m  # radius of BB bullet
    L = p.characteristic_length_m  # characteristic_length_m

    ftype = env.fluid_type
    P = env.pressure_pa  # pressure in Pa
    T = env.temprature_celsius  # temprature in celsius
    g = env.gravity_m_per_s2  # gravitational acceleration
    phi = env.relative_humidity_percent  # relative humidity in percent
    u = env.flow_velocity_m_per_s  # flow velocity in m/s

    v = x[3:6]
    V = np.array([v[0] - u[0], v[1] - u[1], v[2] - u[2]], dtype=np.float64)
    V_norm = np.linalg.norm(V)

    w = x[6:9]
    V_cross_w = np.cross(V, w)

    rho = density_kg_per_m3(T, P, phi, ftype)
    Re = reynolds_number(V_norm, L, T, P, phi, ftype)
    Ma = mach_number(V_norm, T, P, phi, ftype)
    Cd = drag_coefficients_sphere(Re, Ma, ftype)
    Cl = 0.12

    # Force
    # - Gravity force
    Fg = np.array([0.0, 0.0, -m * g], dtype=np.float64)
    # - Drag force
    Fd = -0.5 * Cd * rho * np.pi * R * R * V_norm * V
    # - Magnus force
    Fl = -Cl * ((4.0 * np.pi * R**3) / 3.0) * 2.0 * rho * V_cross_w

    # Define right-hand side of equation of motion
    x_rhs[0:3] = v[0:3]
    x_rhs[3:6] = (Fg + Fd + Fl) / m
    x_rhs[6] = 0.0  # Don't change rotation
    x_rhs[7] = 0.0  # Don't change rotation
    x_rhs[8] = 0.0  # Don't change rotation

    return x_rhs
