# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
import os
import math
from numpy.typing import NDArray
from typing import Union, Tuple
from class_def.environment import Environment
from class_def.particle import Particle
from ivp_functions.main_ivp import solve_ivp_for_bb_bullet
from output_functions.out_file import textout
from output_functions.plot_graph import plot_xy


def compute_ballistics(
    gravity_m_per_s2: Union[float, np.float64],
    fluid_type: str,
    temprature_celsius: Union[float, np.float64],
    pressure_pa: Union[float, np.float64],
    relative_humidity_percent: Union[float, np.float64],
    flow_velocity_m_per_s: NDArray[np.float64],
    mass_kg: Union[float, np.float64],
    radius_m: Union[float, np.float64],
    initial_time_s: Union[float, np.float64],
    initial_position_m: NDArray[np.float64],
    initial_velocity_m_per_s: NDArray[np.float64],
    initial_rotation_rad_per_s: NDArray[np.float64],
    analysis_type: str,
    output_time_step_s: Union[float, np.float64],
    num_of_output: Union[int, np.int32],
    output_folder_path: str,
    is_output_all_figure: bool,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    #    # -- Input parameters start --
    #
    #    # Grobal
    #    gravity_m_per_s2 = 9.80665
    #
    #    # Fluid (e.g. air)
    #    temprature_celsius = 20.0
    #    pressure_pa = 101325.0
    #    relative_humidity_percent = 60.0
    #    fluid_type = "air"
    #    flow_velocity_m_per_s = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    #
    #    # Sphere particle (e.g. BB bullet)
    #    mass_kg = 0.25e-3
    #    radius_m = 3e-3
    #
    #    # Initial values
    #    initial_time_s = 0.0
    #    initial_position_m = np.array([0.0, 0.0, 1.0], dtype=np.float64)
    #    initial_velocity_m_per_s = np.array([84.8528137423857, 0.0, 0.0], dtype=np.float64)
    #    initial_rotation_rad_per_s = np.array([0.0, -2.0 * np.pi * 300, 0.0], dtype=np.float64)
    #
    #    # Analysis type
    #    # "exact": Solve exact rotation decaying
    #    # "approx": Approximate rotation decaying
    #    # "approx_w_solved": Use analytical solution of rotation decaying
    #    analysis_type = "approx"
    #
    #    # -- Input parameters end --

    # Define Environmental characteristic
    environment = Environment(fluid_type, temprature_celsius, pressure_pa, relative_humidity_percent, gravity_m_per_s2, flow_velocity_m_per_s)
    # Define BB-bullet's characteristic
    particle = Particle(mass_kg, radius_m, initial_time_s, initial_position_m, initial_velocity_m_per_s, initial_rotation_rad_per_s)

    # Set output time grid
    start_time_s = initial_time_s
    # output_time_step_s = 0.025
    # num_of_output = 101
    max_time_s = start_time_s + output_time_step_s * (num_of_output - 1)
    time_sequence_s = np.linspace(start_time_s, max_time_s, num_of_output)

    # Solve IVP
    sol_t, sol_y, _ = solve_ivp_for_bb_bullet(analysis_type, time_sequence_s, particle, environment)

    # ----- Output raw data -----
    output_filename = os.path.join(output_folder_path, "out_" + analysis_type + ".txt")
    textout(
        output_filename,
        {
            "# time": sol_t,
            "x": sol_y[0],
            "y": sol_y[1],
            "z": sol_y[2],
            "vx": sol_y[3],
            "vy": sol_y[4],
            "vz": sol_y[5],
            "wx": sol_y[6],
            "wy": sol_y[7],
            "wz": sol_y[8],
        },
    )

    # ----- Output image -----
    x = {"Ballistics": sol_y[0]}
    y = {"Ballistics": sol_y[2]}
    initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
    graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
    x_max = math.ceil(sol_y[0].max() / 10) * 10
    y_max = math.ceil(sol_y[2].max() / 1) * 1
    if y_max < 0.1:
        y_max = 1.0
    plot_xy(
        os.path.join(output_folder_path, "x_z.png"),
        x,
        y,
        x_label="x [m]",
        x_range=(0.0, x_max),
        y_label="height z[m]",
        y_range=(0.0, y_max),
        graph_title=graph_title,
        figure_size_tuple=(6, 3),
        graph_title_size=6,
        legend_loc_tuple=(0.7, 0.8),
    )

    if is_output_all_figure:
        # ----- Output image -----
        x = {"Ballistics": sol_y[1]}
        y = {"Ballistics": sol_y[2]}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[1].max() / 1) * 1
        y_max = math.ceil(sol_y[2].max() / 1) * 1
        if y_max < 0.1:
            y_max = 1.0
        if x_max < 0.1:
            x_max = 1.0
        plot_xy(
            os.path.join(output_folder_path, "y_z.png"),
            x,
            y,
            x_label="horizontal y [m]",
            x_range=(-x_max, x_max),
            y_label="height z[m]",
            y_range=(0.0, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[1]}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil(sol_y[1].max() / 1) * 1
        y_max2 = math.floor(sol_y[1].min() / 1) * 1
        y_max = max(abs(y_max1), abs(y_max2))
        if y_max < 0.1:
            y_max = 1.0
        plot_xy(
            os.path.join(output_folder_path, "x_y.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="horizontal y[m]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[3]}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil(sol_y[3].max() / 100) * 100
        y_max2 = math.floor(sol_y[3].min() / 100) * 100
        y_max = max(abs(y_max1), abs(y_max2))
        if y_max < 0.1:
            y_max = 1.0
        plot_xy(
            os.path.join(output_folder_path, "x_vx.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="velocity vx [m/s]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[4]}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil(sol_y[4].max() / 10) * 10
        y_max2 = math.floor(sol_y[4].min() / 10) * 10
        y_max = max(abs(y_max1), abs(y_max2))

        if y_max < 0.1:
            y_max = 1.0
        plot_xy(
            os.path.join(output_folder_path, "x_vy.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="velocity vy [m/s]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[5]}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil(sol_y[5].max() / 10) * 10
        y_max2 = math.floor(sol_y[5].min() / 10) * 10
        y_max = max(abs(y_max1), abs(y_max2))
        if y_max < 0.1:
            y_max = 1.0

        plot_xy(
            os.path.join(output_folder_path, "x_vz.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="velocity vz [m/s]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": np.sqrt(sol_y[3] ** 2 + sol_y[4] ** 2 + sol_y[5] ** 2)}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max = math.ceil(np.sqrt(sol_y[3] ** 2 + sol_y[4] ** 2 + sol_y[5] ** 2).max() / 100) * 100
        if y_max < 0.1:
            y_max = 1.0

        plot_xy(
            os.path.join(output_folder_path, "x_v.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="velocity |v| [m/s]",
            y_range=(0.0, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[6] / (2.0 * np.pi)}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil((sol_y[6] / (2.0 * np.pi)).max() / 100) * 100
        y_max2 = math.floor((sol_y[6] / (2.0 * np.pi)).min() / 100) * 100
        y_max = max(abs(y_max1), abs(y_max2))
        if y_max < 0.1:
            y_max = 50
        plot_xy(
            os.path.join(output_folder_path, "x_wx.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="rotation wx [rot/s]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[7] / (2.0 * np.pi)}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil((sol_y[7] / (2.0 * np.pi)).max() / 100) * 100
        y_max2 = math.floor((sol_y[7] / (2.0 * np.pi)).min() / 100) * 100
        y_max = max(abs(y_max1), abs(y_max2))
        if y_max < 0.1:
            y_max = 50
        plot_xy(
            os.path.join(output_folder_path, "x_wy.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="rotation wy [rot/s]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_y[8] / (2.0 * np.pi)}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max1 = math.ceil((sol_y[8] / (2.0 * np.pi)).max() / 100) * 100
        y_max2 = math.floor((sol_y[8] / (2.0 * np.pi)).min() / 100) * 100
        y_max = max(abs(y_max1), abs(y_max2))
        if y_max < 0.1:
            y_max = 50
        plot_xy(
            os.path.join(output_folder_path, "x_wz.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="rotation wz [rot/s]",
            y_range=(-y_max, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": np.sqrt(sol_y[6] ** 2 + sol_y[7] ** 2 + sol_y[8] ** 2) / (2.0 * np.pi)}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        y_max = math.ceil((np.sqrt(sol_y[6] ** 2 + sol_y[7] ** 2 + sol_y[8] ** 2) / (2.0 * np.pi)).max() / 100) * 100
        if y_max < 0.1:
            y_max = 50
        plot_xy(
            os.path.join(output_folder_path, "x_w.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, x_max),
            y_label="rotation |w| [rot/s]",
            y_range=(0.0, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

        # ----- Output image -----
        x = {"Ballistics": sol_y[0]}
        y = {"Ballistics": sol_t}
        initial_rotation = particle.initial_rotation_rad_per_s[1] / (2.0 * np.pi)
        graph_title = f"T (deg C) = {temprature_celsius}, P (Pa) = {int(pressure_pa)}, humidity (%) = {relative_humidity_percent}\n BB mass (g) = {mass_kg*1e3:.2f}, initial energy (J) = {particle.initial_kinetic_energy_J:.2f}, initial rotation (rot/s) = {initial_rotation:.2f}"
        x_max = math.ceil(sol_y[0].max() / 10) * 10
        y_max = math.ceil(sol_t.max() / 1) * 1
        if y_max < 0.1:
            y_max = 1

        plot_xy(
            os.path.join(output_folder_path, "x_t.png"),
            x,
            y,
            x_label="x [m]",
            x_range=(0.0, 60),
            y_label="time [s]",
            y_range=(0.0, y_max),
            graph_title=graph_title,
            figure_size_tuple=(6, 3),
            graph_title_size=6,
            legend_loc_tuple=(0.7, 0.8),
        )

    return sol_t, sol_y
